class Versioner(object):
    def get(self):
        pass

    def set(self, version):
        pass

class FileVersioner(Versioner):
    def __init__(self, path='.roetsjbaan-version'):
        self.path = path
        self.version = None
        try:
            with open(self.path, 'r') as f:
                self.version = f.read().strip()
        except IOError:
            pass

    def get(self):
        return self.version

    def set(self, version):
        self.version = version
        with open(self.path, 'wb') as f:
            f.write(self.version)
        return self.version



import os, sys, glob, time, hashlib, inspect
from slugify import slugify
import roetsjbaan.messages as messages
from roetsjbaan.datatypes import SliceableDict

migration_template = '''description = '%s'
hash = '%s'
issue = '%s'

def up():
    pass

def down():
    pass
'''

def module_name(path):
    _, filename = os.path.split(path)
    name, _ = os.path.splitext(filename)
    return name

class MigrationError(Exception):
    def __init__(self, message):
        super(Exception, self).__init__(message)

class Migration(object):
    def __init__(self, module_name):
        self.module = __import__(module_name)
        self.timestamp = int(module_name.split('_')[0])
        self.hash = self.module.hash
        self.description = self.module.description
        self.issue = self.module.issue

    def up(self, inject):
        args = inspect.getargspec(self.module.up).args
        return self.module.up(*[inject[arg] for arg in args])

    def down(self, inject):
        args = inspect.getargspec(self.module.down).args
        return self.module.down(*[inject[arg] for arg in args])

class InitialMigration(Migration):
    def __init__(self):
        self.timestamp = 0
        self.hash = '0' * 16
        self.description = 'Original state'
        self.issue = 'None'

    def up(self, inject):
        pass

    def down(self, inject):
        return False

class Migrator(object):
    def __init__(self, versioner, directory='migrations', inject={}):
        self.versioner = versioner
        self.directory = directory
        self.migrations = SliceableDict()
        self.inject = inject

        if directory not in sys.path:
            sys.path.append(directory)

        self.reload()

    def up(self, to=None):
        '''
        Migrate up to the latest migration. If <to> is specified, migrate up to
        <to> and stop.
        '''
        current = self.versioner.get()
        if to:
            to = self.unique(to).hash
        else:
            # Go up to the most recent migration
            to = self.all()[-1].hash

        migrations = self.migrations[current:to, 1]
        frm = migrations[0]

        if len(migrations) < 2:
            raise MigrationError('No migrations left to do.')

        for to in migrations[1:]:
            yield messages.Migrating(frm, to, 'up')
            to.up(self.inject)
            self.versioner.set(to.hash)
            frm = to

    def down(self, to=None):
        '''
        Migrate down to <to>. if <to> is not specified, migrate down one
        migration
        '''
        current = self.versioner.get()

        if to:
            to = self.unique(to).hash
            migrations = list(reversed(self.migrations[to:current, 1]))
            # pack it into (to => from pairs).
            from_to = zip(migrations[:-1], migrations[1:])
        else:
            migrations = self.migrations[current::-1][:2]

            if len(migrations) < 2:
                raise MigrationError('No migrations left to undo.')

            from_to = [migrations]

        for frm, to in from_to:
            yield messages.Migrating(frm, to, 'down')
            frm.down(self.inject)
            self.versioner.set(to.hash)

    def create(self, description, issue=None):
        timestamp = int(time.time())
        slug = slugify(unicode(description)).replace('-', '_')
        hash = hashlib.sha1('%d%s%s' % (timestamp, description, str(issue))).hexdigest()
        name = '%d_%s_%s.py' % (timestamp, hash[:8], slug)

        with open(os.path.join(self.directory, name), 'wb') as f:
            f.write(migration_template % (description, hash, str(issue)))


        return name

    def find(self, hash):
        return [
            m for m in self.migrations.values()
            if m.hash.startswith(hash)
        ]

    def unique(self, hash):
        hits = self.find(hash)

        if len(hits) > 1:
            raise IndexError('Multiple hashes found with prefix ' + hash)

        if not hits:
            raise IndexError('No hash found with prefix ' + hash)

        return hits[0]

    def reload(self):
        migrations = []
        for f in glob.glob(os.path.join(self.directory, '*.py')):
            m = Migration(module_name(f))
            migrations.append((m.hash, m))

        initial = InitialMigration()
        migrations.append((initial.hash, initial))
        migrations.sort(key=lambda m: int(m[1].timestamp))
        self.migrations = SliceableDict(migrations)

    def __iter__(self):
        for m in self.migrations.values():
            yield m

    def all(self):
        return self.migrations.values()



import collections

class SliceableDict(collections.OrderedDict):
    def __getitem__(self, key):
        stop_offset = 0

        if isinstance(key, tuple):
            key, stop_offset = key

        if isinstance(key, slice):
            return self.values()[self.__calculate_slice(key, stop_offset)]

        return super(SliceableDict, self).__getitem__(key)

    def __calculate_slice(self, key, stop_offset=0):
        start, stop, step = key.start, key.stop, key.step

        if start:
            start = next(
                i for i, (k, v) in enumerate(self.items())
                if k == start
            )

        if stop:
            stop = next(
                i for i, (k, v) in enumerate(self.items())
                if k == stop
            ) + stop_offset


        return slice(start, stop, step)



from setuptools import setup

setup(
    name='roetsjbaan',
    version='0.1.4',
    description='A minimal, lightweight migration system',
    url='https://github.com/mivdnber/roetsjbaan',
    download_url='https://github.com/mivdnber/roetsjbaan/tarball/0.1.4',
    author='Michiel Van den Berghe',
    author_email='michiel.vdb@gmail.com',
    license='MIT',
    packages=['roetsjbaan'],
    keywords=['migration', 'database'],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'roetsj=roetsjbaan.main:main'
        ]
    },
    install_requires=[
        'slugify==0.0.1',
        'tabulate==0.7.5'
    ]
)



import argparse, sys, time, os
from tabulate import tabulate
import imp

import roetsjbaan
import roetsjbaan.messages as messages

sys.path.append(os.getcwd())
f, path, desc = imp.find_module('roetsjfile', [os.getcwd()])
roetsjfile = imp.load_module('roetsjfile', f, path, desc)

migrator = roetsjbaan.Migrator(roetsjfile.versioner,
    inject=roetsjfile.inject if hasattr(roetsjfile, 'inject') else {},
    directory=roetsjfile.directory if hasattr(roetsjfile, 'directory') else 'migrations'
)

migrator.reload()

def list(args):
    '''
    Shows a list of all migrations.
    '''
    print 'Migrations in %s:' % migrator.directory
    print pretty_migrations(migrator, long=args.long)
    if not migrator.versioner.get():
        print "Warning! I don't know the current version. Please set it with roetsj set."

def new(args):
    '''
    Create a migration with a given description and an optional issue number
    '''
    fn = migrator.create(args.description, args.issue)
    print 'Migration created as %s' % fn

def set(args):
    '''
    Set the current version. You shouldn't need this unless you screw up a
    migration
    '''
    hits = migrator.find(args.hash)
    if len(hits) > 1:
        print 'More than one migration corresponds to hash %s:' % args.hash
    else:
        migrator.versioner.set(hits[0].hash)
        print 'Current version set to:'
    print pretty_migrations(hits)

def up(args):
    '''
    Execute all pending migrations. You can specify op to which migration you
    wish to migrate with the -t <hash> or --to <hash> options.
    '''
    if not migrator.versioner.get():
        print 'Sorry, the current version is not set. Please set it with roetsj set.'
        return
    try:
        for message in migrator.up(args.to):
            print 'up', pretty_migrations([message.new], headers=[], format='plain')
    except roetsjbaan.MigrationError as e:
        print e.message

def down(args):
    '''
    Undo the current migration. You can undo multiple migrations at once by
    specifying the -t <hash> or --to <hash> options.
    '''
    if not migrator.versioner.get():
        print 'Sorry, the current version is not set. Please set it with roetsj set.'
        return
    try:
        for message in migrator.down(args.to):
            print 'down', pretty_migrations([message.old], headers=[], format='plain')
    except roetsjbaan.MigrationError as e:
        print e.message

def pretty_migrations(migrations, long=False, headers=['', 'Date', 'Hash', 'Description', 'Issue #'], format='simple'):
    '''
    Formats a list of migrations into a sexy ASCII table
    '''
    data = [
        [
            '>' if migrator.versioner.get() == m.hash else ' ',
            time.strftime('%Y-%m-%d', time.gmtime(m.timestamp)),
            m.hash[:len(m.hash) if long else 8], m.description, m.issue
        ]
        for m in migrations
    ]
    return tabulate(reversed(data), headers=headers, tablefmt=format)

def main():
    parser = argparse.ArgumentParser(description='Migrate stuff.')
    subparsers = parser.add_subparsers()

    new_parser = subparsers.add_parser('new', description=new.__doc__)
    new_parser.add_argument('description')
    new_parser.add_argument('--issue')
    new_parser.set_defaults(func=new)

    set_parser = subparsers.add_parser('set')
    set_parser.add_argument('hash')
    set_parser.set_defaults(func=set)

    list_parser = subparsers.add_parser('list', description=list.__doc__)
    list_parser.add_argument('-l', '--long', action='store_true', help='''
        Show the entire migration hash
    ''')
    list_parser.set_defaults(func=list)

    up_parser = subparsers.add_parser('up', description=up.__doc__)
    up_parser.add_argument('-t', '--to', type=str, help='''
        Hash of the migration up to which you wish to upgrade. Defaults to all
        more recent migrations.

    ''')
    up_parser.set_defaults(func=up)

    down_parser = subparsers.add_parser('down', description=down.__doc__)
    down_parser.add_argument('-t', '--to', type=str, help='''
        Hash of the migration up to which you wish to downgrade. Defaults to the
        previous migration.
    ''')
    down_parser.set_defaults(func=down)

    args = parser.parse_args()
    args.func(args)
